from typing import Any, Dict, List, Literal

import pandas as pd
from prefect import Flow, task

from viadot.tasks import AzureDataLakeUpload
from viadot.tasks.genesys import GenesysToCSV, GenesysToDF
from viadot.task_utils import (
    add_ingestion_metadata_task,
    df_to_csv,
    df_to_parquet,
    adls_bulk_upload,
)


@task
def add_timestamp(files_names: List = None, sep: str = None) -> None:
    """Add new column _viadot_downloaded_at_utc into every genesys file.

    Args:
        files_names (List, optional): All file names of downloaded files. Defaults to None.
        sep (str, optional): Separator in csv file. Defaults to None.
    """
    for file in files_names:
        df = pd.read_csv(file, sep=sep)
        df_updated = add_ingestion_metadata_task.run(df)
        df_updated.to_csv(file, index=False, sep=sep)


class GenesysToADLS(Flow):
    def __init__(
        self,
        name: str,
        view_type: str = "queue_performance_detail_view",
        view_type_time_sleep: int = 80,
        media_type_list: List[str] = None,
        queueIds_list: List[str] = None,
        data_to_post_str: str = None,
        start_date: str = None,
        end_date: str = None,
        days_interval: int = 1,
        sep: str = "\t",
        environment: str = None,
        schedule_id: str = None,
        report_url: str = None,
        report_columns: List[str] = None,
        local_file_path: str = None,
        adls_file_path: str = None,
        overwrite_adls: bool = True,
        adls_sp_credentials_secret: str = None,
        credentials_genesys: Dict[str, Any] = None,
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any]
    ):
        """
        Genesys flow that generates CSV files and upload them to ADLS.

        Args:
            name (str): The name of the Flow.
            view_type (str, optional): The type of view export job to be created. Defaults to "queue_performance_detail_view".
            view_type_time_sleep (int, optional): Waiting time to retrieve data from Genesys API. Defaults to 80.
            media_type_list (List[str], optional): List of specific media types. Defaults to None.
            queueIds_list (List[str], optional): List of specific queues ids. Defaults to None.
            data_to_post_str (str, optional): String template to generate json body. Defaults to None.
            start_date (str, optional): Start date of the report. Defaults to None.
            end_date (str, optional): End date of the report. Defaults to None.
            days_interval (int, optional): How many days report should include. Defaults to 1.
            sep (str, optional): Separator in csv file. Defaults to "\t".
            environment (str, optional): Adress of host server. Defaults to None than will be used enviroment
            from credentials.
            schedule_id (str, optional): The ID of report. Defaults to None.
            report_url (str, optional): The url of report generated in json response. Defaults to None.
            report_columns (List[str], optional): List of exisiting column in report. Defaults to None.
            local_file_path (str, optional): The local path from which to upload the file(s). Defaults to None.
            adls_file_path (str, optional): The destination path at ADLS. Defaults to None.
            overwrite_adls (bool, optional): Whether to overwrite files in the data lake. Defaults to True.
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET). Defaults to None.
            credentials(dict, optional): Credentials for the genesys api. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        # GenesysToCSV
        self.flow_name = name
        self.view_type = view_type
        self.view_type_time_sleep = view_type_time_sleep
        self.media_type_list = media_type_list
        self.queueIds_list = queueIds_list
        self.data_to_post = data_to_post_str
        self.environment = environment
        self.schedule_id = schedule_id
        self.report_url = report_url
        self.report_columns = report_columns
        self.start_date = start_date
        self.end_date = end_date
        self.days_interval = days_interval
        self.sep = sep
        self.timeout = timeout

        # AzureDataLake
        self.local_file_path = local_file_path
        self.adls_file_path = adls_file_path
        self.overwrite_adls = overwrite_adls
        self.adls_sp_credentials_secret = adls_sp_credentials_secret
        self.credentials_genesys = credentials_genesys

        super().__init__(*args, name=self.flow_name, **kwargs)

        self.gen_flow()

    def gen_flow(self) -> Flow:

        to_csv = GenesysToCSV(timeout=self.timeout)

        if self.view_type == "queue_performance_detail_view":
            file_names = to_csv.bind(
                view_type=self.view_type,
                view_type_time_sleep=self.view_type_time_sleep,
                media_type_list=self.media_type_list,
                queueIds_list=self.queueIds_list,
                data_to_post_str=self.data_to_post,
                start_date=self.start_date,
                end_date=self.end_date,
                days_interval=self.days_interval,
                environment=self.environment,
                credentials_genesys=self.credentials_genesys,
                flow=self,
            )
        elif self.view_type in [
            "agent_performance_summary_view",
            "agent_status_summary_view",
        ]:
            file_names = to_csv.bind(
                view_type=self.view_type,
                view_type_time_sleep=self.view_type_time_sleep,
                media_type_list=self.media_type_list,
                queueIds_list=[""],
                data_to_post_str=self.data_to_post,
                start_date=self.start_date,
                environment=self.environment,
                credentials_genesys=self.credentials_genesys,
                flow=self,
            )

        add_timestamp.bind(file_names, sep=self.sep, flow=self)

        adls_bulk_upload.bind(
            file_names=file_names,
            adls_file_path=self.adls_file_path,
            adls_sp_credentials_secret=self.adls_sp_credentials_secret,
            timeout=self.timeout,
            flow=self,
        )

        add_timestamp.set_upstream(file_names, flow=self)
        adls_bulk_upload.set_upstream(add_timestamp, flow=self)


class GenesysReportToADLS(Flow):
    def __init__(
        self,
        name: str,
        columns: List[str] = None,
        vault_name: str = None,
        local_file_path: str = None,
        output_file_extension: str = ".csv",
        sep: str = "\t",
        adls_file_path: str = None,
        if_exists: Literal["replace", "append", "delete"] = "replace",
        overwrite_adls: bool = True,
        adls_sp_credentials_secret: str = None,
        credentials_secret: str = None,
        schedule_id: str = None,
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any]
    ):
        """
        Flow for downloading data from genesys API to Azure Data Lake.

        Args:
            name (str): The name of the flow.
            columns (List[str], optional): The report fields. Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secrets. Defaults to None.
            local_file_path (str, optional): Local destination path. Defaults to None.
            output_file_extension (str, optional): Output file extension - to allow selection of.csv for data
            which is not easy to handle with parquet. Defaults to ".csv".
            sep (str, optional): The separator used in csv. Defaults to "\t".
            adls_file_path (str, optional): Azure Data Lake destination file path. Defaults to None.
            if_exists (str, optional): What to do if the file exists. Defaults to "replace".
            overwrite_adls (bool, optional): Whether to overwrite files in the lake. Defaults to True.
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
            Defaults to None.
            credentials_secret (str, optional): The name of the Azure Key Vault secret for Genesys project. Defaults to None.
            schedule_id (str, optional): ID of the schedule report job. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """

        self.name = name
        self.columns = columns
        self.vault_name = vault_name
        self.local_file_path = local_file_path
        self.output_file_extension = output_file_extension
        self.sep = sep
        self.if_exists = if_exists
        self.adls_file_path = adls_file_path
        self.overwrite_adls = overwrite_adls
        self.adls_sp_credentials_secret = adls_sp_credentials_secret
        self.credentials_secret = credentials_secret
        self.if_exsists = if_exists
        self.schedule_id = schedule_id
        self.timeout = timeout

        super().__init__(*args, name=name, **kwargs)

        self.gen_flow()

    def gen_flow(self) -> Flow:

        genesys_report = GenesysToDF(timeout=self.timeout)

        df = genesys_report.bind(
            report_columns=self.columns,
            schedule_id=self.schedule_id,
            credentials_genesys=self.credentials_secret,
            flow=self,
        )
        df_with_metadata = add_ingestion_metadata_task.bind(df, flow=self)

        if self.output_file_extension == ".parquet":
            df_to_file = df_to_parquet.bind(
                df=df_with_metadata,
                path=self.local_file_path,
                if_exists=self.if_exsists,
                flow=self,
            )
        else:
            df_to_file = df_to_csv.bind(
                df=df_with_metadata,
                path=self.local_file_path,
                sep=self.sep,
                if_exists=self.if_exsists,
                flow=self,
            )

        file_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        file_to_adls_task.bind(
            from_path=self.local_file_path,
            to_path=self.adls_file_path,
            overwrite=self.overwrite_adls,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        df_with_metadata.set_upstream(df, flow=self)
        df_to_file.set_upstream(df_with_metadata, flow=self)
        file_to_adls_task.set_upstream(df_to_file, flow=self)



import json
from typing import Any, Dict, List

import pandas as pd
from prefect import Task
from prefect.utilities import logging
from prefect.utilities.tasks import defaults_from_attrs

from ..exceptions import DBDataAccessError
from ..sources import BigQuery
from .azure_key_vault import AzureKeyVaultSecret

logger = logging.get_logger()


class BigQueryToDF(Task):
    """
    The task for querying BigQuery and saving data as the data frame.
    """

    def __init__(
        self,
        dataset_name: str = None,
        table_name: str = None,
        start_date: str = None,
        end_date: str = None,
        date_column_name: str = "date",
        credentials_key: str = None,
        credentials_secret: str = None,
        vault_name: str = None,
        timeout: int = 3600,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """
        Initialize BigQueryToDF object. For querying on database - dataset and table name is required.
        The name of the project is taken from the config/credential json file so there is no need to enter its name.

        There are 3 cases:
            If start_date and end_date are not None - all data from the start date to the end date will be retrieved.
            If start_date and end_date are left as default (None) - the data is pulled till "yesterday" (current date -1)
            If the column that looks like a date does not exist in the table, get all the data from the table.

        Args:
            dataset_name (str, optional): Dataset name. Defaults to None.
            table_name (str, optional): Table name. Defaults to None.
            date_column_name (str, optional): The query is based on a date, the user can provide the name
            of the date columnn if it is different than "date". If the user-specified column does not exist,
            all data will be retrieved from the table. Defaults to "date".
            start_date (str, optional): A query parameter to pass start date e.g. "2022-01-01". Defaults to None.
            end_date (str, optional): A query parameter to pass end date e.g. "2022-01-01". Defaults to None.
            credentials_key (str, optional): Credential key to dictionary where details are stored (local config).
            credentials can be generated as key for User Principal inside a BigQuery project. Defaults to None.
            credentials_secret (str, optional): The name of the Azure Key Vault secret for Bigquery project. Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secrets. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        self.dataset_name = dataset_name
        self.table_name = table_name
        self.start_date = start_date
        self.end_date = end_date
        self.date_column_name = date_column_name
        self.credentials_key = credentials_key
        self.credentials_secret = credentials_secret
        self.vault_name = vault_name

        super().__init__(
            name="bigquery_to_df",
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self):
        """Download BigQuery data to a DF"""
        super().__call__(self)

    @defaults_from_attrs(
        "dataset_name",
        "table_name",
        "date_column_name",
        "start_date",
        "end_date",
        "credentials_key",
        "credentials_secret",
        "vault_name",
    )
    def run(
        self,
        dataset_name: str = None,
        table_name: str = None,
        date_column_name: str = "date",
        start_date: str = None,
        end_date: str = None,
        credentials_key: str = None,
        credentials_secret: str = None,
        vault_name: str = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        credentials = None
        if credentials_secret:
            credentials_str = AzureKeyVaultSecret(
                credentials_secret, vault_name=vault_name
            ).run()
            credentials = json.loads(credentials_str)

        bigquery = BigQuery(credentials_key=credentials_key, credentials=credentials)
        project = bigquery.get_project_id()
        try:
            table_columns = bigquery.list_columns(
                dataset_name=dataset_name, table_name=table_name
            )
            if date_column_name not in table_columns:
                logger.warning(
                    f"'{date_column_name}' column is not recognized. Downloading all the data from '{table_name}'."
                )
                query = f"SELECT * FROM `{project}.{dataset_name}.{table_name}`"
                df = bigquery.query_to_df(query)
            else:
                if start_date is not None and end_date is not None:
                    query = f"""SELECT * FROM `{project}.{dataset_name}.{table_name}` 
                    where {date_column_name} between PARSE_DATE("%Y-%m-%d", "{start_date}") and PARSE_DATE("%Y-%m-%d", "{end_date}") 
                    order by {date_column_name} desc"""
                else:
                    query = f"""SELECT * FROM `{project}.{dataset_name}.{table_name}` 
                    where {date_column_name} < CURRENT_DATE() 
                    order by {date_column_name} desc"""

                df = bigquery.query_to_df(query)
                logger.info(
                    f"Downloaded the data from the '{table_name}' table into the data frame."
                )

        except DBDataAccessError:
            logger.warning(
                f"Table name '{table_name}' or dataset '{dataset_name}' not recognized. Returning empty data frame."
            )
            df = pd.DataFrame()
        return df



from viadot.tasks import GenesysToCSV, GenesysToDF

import os
from unittest import mock

import pandas as pd
import pytest

MEDIA_TYPE_LIST = ["callback"]

QUEUEIDS_LIST = ["780807e6-83b9-44be-aff0-a41c37fab004"]

DATA_TO_POST_STR = """{
    "name": f"QUEUE_PERFORMANCE_DETAIL_VIEW_{media}",
    "timeZone": "UTC",
    "exportFormat": "CSV",
    "interval": f"{end_date}T23:00:00/{start_date}T23:00:00",
    "period": "PT30M",
    "viewType": f"QUEUE_PERFORMANCE_DETAIL_VIEW",
    "filter": {"mediaTypes": [f"{media}"], "queueIds": [f"{queueid}"], "directions":["inbound"],},
    "read": True,
    "locale": "en-us",
    "hasFormatDurations": False,
    "hasSplitFilters": True,
    "excludeEmptyRows": True,
    "hasSplitByMedia": True,
    "hasSummaryRow": True,
    "csvDelimiter": "COMMA",
    "hasCustomParticipantAttributes": True,
    "recipientEmails": [],
    }"""

d = {"country": [1, 2], "sales": [3, 4]}
df = pd.DataFrame(data=d)

genesys_to_csv = GenesysToCSV()
genests_to_df = GenesysToDF()


def test_genesys_to_csv():
    with mock.patch.object(
        genesys_to_csv,
        "run",
        return_value=["V_D_PROD_FB_QUEUE_CALLBACK.csv", "V_D_PROD_FB_QUEUE_CHAT.csv"],
    ) as mock_method:

        files_name_list = genesys_to_csv.run(
            media_type_list=MEDIA_TYPE_LIST,
            queueIds_list=QUEUEIDS_LIST,
            data_to_post_str=DATA_TO_POST_STR,
        )
        assert isinstance(files_name_list, list)


def test_genests_to_df():
    with mock.patch.object(
        genests_to_df,
        "run",
        return_value=df,
    ) as mock_method:

        final_df = genests_to_df.run(
            report_url="https://apps.mypurecloud.de/platform/api/v2/downloads/62631b245c69f0c8",
        )
        assert isinstance(final_df, pd.DataFrame)



import logging
import os

import pandas as pd
import pytest

from viadot.signals import SKIP
from viadot.utils import check_if_empty_file, gen_bulk_insert_query_from_df

EMPTY_CSV_PATH = "empty.csv"
EMPTY_PARQUET_PATH = "empty.parquet"


def test_single_quotes_inside():
    TEST_VALUE = "a'b"
    df1 = pd.DataFrame(
        {
            "a": [
                TEST_VALUE,
            ],
            "b": ["a"],
        }
    )
    test_insert_query = gen_bulk_insert_query_from_df(
        df1, table_fqn="test_schema.test_table"
    )
    TEST_VALUE_ESCAPED = "'a''b'"
    assert (
        test_insert_query
        == f"""INSERT INTO test_schema.test_table (a, b)

VALUES ({TEST_VALUE_ESCAPED}, 'a')"""
    ), test_insert_query


def test_single_quotes_outside():
    TEST_VALUE = "'a'"
    df1 = pd.DataFrame(
        {
            "a": [
                TEST_VALUE,
            ],
            "b": ["b"],
        }
    )
    test_insert_query = gen_bulk_insert_query_from_df(
        df1, table_fqn="test_schema.test_table"
    )
    TEST_VALUE_ESCAPED = "'''a'''"
    assert (
        test_insert_query
        == f"""INSERT INTO test_schema.test_table (a, b)

VALUES ({TEST_VALUE_ESCAPED}, 'b')"""
    ), test_insert_query


def test_double_quotes_inside():
    TEST_VALUE = 'a "b"'
    df1 = pd.DataFrame(
        {
            "a": [
                TEST_VALUE,
            ],
            "b": ["c"],
        }
    )
    test_insert_query = gen_bulk_insert_query_from_df(
        df1, table_fqn="test_schema.test_table"
    )
    TEST_VALUE_ESCAPED = """'a "b"'"""
    assert (
        test_insert_query
        == f"""INSERT INTO test_schema.test_table (a, b)

VALUES ({TEST_VALUE_ESCAPED}, 'c')"""
    ), test_insert_query


def test_check_if_empty_file_csv(caplog):
    with open(EMPTY_CSV_PATH, "w"):
        pass

    with caplog.at_level(logging.WARNING):
        check_if_empty_file(path=EMPTY_CSV_PATH, if_empty="warn")
        assert f"Input file - '{EMPTY_CSV_PATH}' is empty." in caplog.text
    with pytest.raises(ValueError):
        check_if_empty_file(path=EMPTY_CSV_PATH, if_empty="fail")
    with pytest.raises(SKIP):
        check_if_empty_file(path=EMPTY_CSV_PATH, if_empty="skip")

    os.remove(EMPTY_CSV_PATH)


def test_check_if_empty_file_parquet(caplog):
    with open(EMPTY_PARQUET_PATH, "w"):
        pass

    with caplog.at_level(logging.WARNING):
        check_if_empty_file(path=EMPTY_PARQUET_PATH, if_empty="warn")
        assert f"Input file - '{EMPTY_PARQUET_PATH}' is empty." in caplog.text
    with pytest.raises(ValueError):
        check_if_empty_file(path=EMPTY_PARQUET_PATH, if_empty="fail")
    with pytest.raises(SKIP):
        check_if_empty_file(path=EMPTY_PARQUET_PATH, if_empty="skip")

    os.remove(EMPTY_PARQUET_PATH)


def test_check_if_empty_file_no_data(caplog):
    df = pd.DataFrame({"col1": []})
    df.to_parquet(EMPTY_PARQUET_PATH)
    with caplog.at_level(logging.WARNING):
        check_if_empty_file(path=EMPTY_PARQUET_PATH, if_empty="warn")
        assert f"Input file - '{EMPTY_PARQUET_PATH}' is empty." not in caplog.text



from .azure_blob_storage import AzureBlobStorage
from .azure_data_lake import AzureDataLake
from .azure_sql import AzureSQL
from .bigquery import BigQuery
from .cloud_for_customers import CloudForCustomers
from .outlook import Outlook
from .salesforce import Salesforce
from .sharepoint import Sharepoint
from .supermetrics import Supermetrics
from .genesys import Genesys
from .sftp import SftpConnector


try:
    from .sap_rfc import SAPRFC
except ImportError:
    pass

from .duckdb import DuckDB
from .epicor import Epicor
from .sql_server import SQLServer
from .sqlite import SQLite
from .mindful import Mindful

# APIS
from .uk_carbon_intensity import UKCarbonIntensity



import pandas as pd

from viadot.sources import BigQuery

BIGQ = BigQuery(credentials_key="BIGQUERY_TESTS")


def test_list_project():
    project = BIGQ.get_project_id()
    assert project == "manifest-geode-341308"


def test_list_datasets():
    datasets = list(BIGQ.list_datasets())
    assert datasets == ["manigeo", "official_empty"]


def test_list_tables():
    datasets = BIGQ.list_datasets()
    tables = list(BIGQ.list_tables(datasets[0]))
    assert tables == ["space", "test_data", "manigeo_tab"]


def test_query_is_df():
    query = """
            SELECT name, SUM(number) AS total
            FROM `bigquery-public-data.usa_names.usa_1910_2013`
            GROUP BY name, gender
            ORDER BY total DESC 
            LIMIT 4
            """
    df = BIGQ.query_to_df(query)

    assert isinstance(df, pd.DataFrame)


def test_query():
    query = """
            SELECT name, SUM(number) AS total
            FROM `bigquery-public-data.usa_names.usa_1910_2013`
            GROUP BY name, gender
            ORDER BY total DESC 
            LIMIT 4
            """
    df = BIGQ.query_to_df(query)
    total_received = df["total"].values

    assert total_received == [4924235, 4818746, 4703680, 4280040]



import time
from typing import Any, Dict, List, Literal

import pandas as pd
import numpy as np
import prefect
from prefect import Task
from prefect.utilities import logging
from prefect.utilities.tasks import defaults_from_attrs

from ..exceptions import APIError
from ..sources import Genesys

logger = logging.get_logger()


class GenesysToCSV(Task):
    def __init__(
        self,
        report_name: str = "genesys_to_csv",
        view_type: str = "queue_performance_detail_view",
        media_type_list: List[str] = None,
        queueIds_list: List[str] = None,
        data_to_post_str: str = None,
        start_date: str = None,
        end_date: str = None,
        days_interval: int = 1,
        environment: str = None,
        schedule_id: str = None,
        report_url: str = None,
        report_columns: List[str] = None,
        credentials_genesys: Dict[str, Any] = None,
        timeout: int = 3600,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """Task for downloading data from Genesys API to CSV.

        Args:
            report_name (str, optional): The name of this task. Defaults to a general name 'genesys_to_csv'.
            view_type (str, optional): The type of view export job to be created. Defaults to "queue_performance_detail_view".
            media_type_list (List[str], optional): List of specific media types. Defaults to None.
            queueIds_list (List[str], optional): List of specific queues ids. Defaults to None.
            data_to_post_str (str, optional): String template to generate json body. Defaults to None.
            credentials_genesys (Dict[str, Any], optional): Credentials to connect with Genesys API containing CLIENT_ID. Defaults to None.
            start_date (str, optional): Start date of the report. Defaults to None.
            end_date (str, optional): End date of the report. Defaults to None.
            days_interval (int, optional): How many days report should include. Defaults to 1.
            environment (str, optional): Adress of host server. Defaults to None than will be used enviroment
            from credentials.
            schedule_id (str, optional): The ID of report. Defaults to None.
            report_url (str, optional): The url of report generated in json response. Defaults to None.
            report_columns (List[str], optional): List of exisiting column in report. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """

        self.logger = prefect.context.get("logger")
        self.schedule_id = schedule_id
        self.report_name = report_name
        self.view_type = view_type
        self.environment = environment
        self.report_url = report_url
        self.report_columns = report_columns
        self.media_type_list = media_type_list
        self.queueIds_list = queueIds_list
        self.data_to_post_str = data_to_post_str
        self.start_date = start_date
        self.end_date = end_date
        self.days_interval = days_interval
        self.credentials_genesys = credentials_genesys

        super().__init__(
            name=self.report_name,
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Download Genesys data to CSV"""
        return super().__call__(*args, **kwargs)

    @defaults_from_attrs(
        "report_name",
        "view_type",
        "environment",
        "schedule_id",
        "report_url",
        "media_type_list",
        "queueIds_list",
        "data_to_post_str",
        "start_date",
        "end_date",
        "report_columns",
        "days_interval",
        "credentials_genesys",
    )
    def run(
        self,
        report_name: str = None,
        view_type: str = "queue_performance_detail_view",
        view_type_time_sleep: int = 80,
        environment: str = None,
        schedule_id: str = None,
        report_url: str = None,
        media_type_list: List[str] = None,
        queueIds_list: List[str] = None,
        data_to_post_str: str = None,
        start_date: str = None,
        end_date: str = None,
        report_columns: List[str] = None,
        days_interval: int = None,
        credentials_genesys: Dict[str, Any] = None,
    ) -> List[str]:
        """
        Task for downloading data from the Genesys API to DF.

        Args:
            report_name (str, optional): The name of this task. Defaults to a general name 'genesys_to_csv'.
            view_type (str, optional): The type of view export job to be created. Defaults to "queue_performance_detail_view".
            view_type_time_sleep (int, optional): Waiting time to retrieve data from Genesys API. Defaults to 80.
            media_type_list (List[str], optional): List of specific media types. Defaults to None.
            queueIds_list (List[str], optional): List of specific queues ids. Defaults to None.
            data_to_post_str (str, optional): String template to generate json body. Defaults to None.
            credentials_genesys (Dict[str, Any], optional): Credentials to connect with Genesys API containing CLIENT_ID. Defaults to None.
            start_date (str, optional): Start date of the report. Defaults to None.
            end_date (str, optional): End date of the report. Defaults to None.
            days_interval (int, optional): How many days report should include. Defaults to 1.
            environment (str, optional): Adress of host server. Defaults to None than will be used enviroment
            from credentials.
            schedule_id (str, optional): The ID of report. Defaults to None.
            report_url (str, optional): The url of report generated in json response. Defaults to None.
            report_columns (List[str], optional): List of exisiting column in report. Defaults to None.

        Returns:
            List[str]: List of file names.
        """

        genesys = Genesys(
            report_name=report_name,
            view_type=view_type,
            media_type_list=media_type_list,
            queueIds_list=queueIds_list,
            data_to_post_str=data_to_post_str,
            credentials_genesys=credentials_genesys,
            start_date=start_date,
            end_date=end_date,
            days_interval=days_interval,
            environment=environment,
            schedule_id=schedule_id,
            report_url=report_url,
            report_columns=report_columns,
        )

        genesys.genesys_generate_body()
        genesys.genesys_generate_exports()

        if view_type == "queue_performance_detail_view":
            logger.info(
                f"Waiting {view_type_time_sleep} seconds for caching data in Genesys database."
            )
            # sleep time to allow Genesys generate all exports
            time.sleep(view_type_time_sleep)
            # in order to wait for API POST request add it
            timeout_start = time.time()
            # 30 seconds timeout is minimal but for safety added 60.
            timeout = timeout_start + 60
            # while loop with timeout
            while time.time() < timeout:

                try:
                    genesys.get_reporting_exports_data()
                    urls = [col for col in np.array(genesys.report_data).T][1]
                    if None in urls:
                        logger.warning("Found None object in list of urls.")
                    else:
                        break
                except TypeError:
                    pass

                # There is a need to clear a list before repeating try statement.
                genesys.report_data.clear()

        elif view_type in [
            "agent_performance_summary_view",
            "agent_status_summary_view",
        ]:
            logger.info(
                f"Waiting for getting data in Genesys database ({view_type_time_sleep} seconds)."
            )
            time.sleep(view_type_time_sleep)

            genesys.get_reporting_exports_data()
        # print(genesys.report_data)
        failed = [col for col in np.array(genesys.report_data).T][-1]

        if "FAILED" in failed and "COMPLETED" in failed:
            logger.warning("Some reports failed.")

        if len(genesys.report_data) == 0 or np.unique(failed)[0] == "FAILED":
            genesys.delete_all_reporting_exports()
            logger.warning(f"All existing reports were delted.")
            raise APIError("No exporting reports were generated.")
        elif not None in [col for col in np.array(genesys.report_data).T][1]:
            logger.info("Downloaded the data from the Genesys into the CSV.")
        else:
            logger.info("Succesfully loaded all exports.")

        file_names = genesys.download_all_reporting_exports()
        logger.info("Downloaded the data from the Genesys into the CSV.")
        # in order to wait for API GET request call it
        logger.info("Waiting for caching data in Genesys database.")
        time.sleep(20)
        genesys.delete_all_reporting_exports()
        logger.info(f"All existing reports were delted.")

        return file_names


class GenesysToDF(Task):
    def __init__(
        self,
        report_name: str = None,
        environment: str = None,
        schedule_id: str = None,
        report_url: str = None,
        report_columns: List[str] = None,
        credentials_genesys: Dict[str, Any] = None,
        timeout: int = 3600,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):

        self.logger = prefect.context.get("logger")
        self.schedule_id = schedule_id
        self.report_name = report_name
        self.environment = environment
        self.report_url = report_url
        self.report_columns = report_columns
        self.credentials_genesys = credentials_genesys

        super().__init__(
            name="genesys_to_df",
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Download Genesys data to DF"""
        return super().__call__(*args, **kwargs)

    @defaults_from_attrs(
        "report_name",
        "environment",
        "schedule_id",
        "report_url",
        "report_columns",
        "credentials_genesys",
    )
    def run(
        self,
        report_name: str = None,
        environment: str = None,
        schedule_id: str = None,
        report_url: str = None,
        report_columns: List[str] = None,
        credentials_genesys: Dict[str, Any] = None,
    ) -> pd.DataFrame:
        """
        Task for downloading data from the Genesys API to DF.

        Args:
            report_name (str, optional): Name of the report. Defaults to None.
            environment (str, optional): Adress of host server. Defaults to None than will be used enviroment
            from credentials.
            schedule_id (str, optional): The ID of report. Defaults to None.
            report_url (str, optional): The url of report generated in json response. Defaults to None.
            report_columns (List[str], optional): List of exisiting column in report. Defaults to None.
            credentials_genesys (Dict[str, Any], optional): Credentials to connect with Genesys API containing CLIENT_ID. Defaults to None.

        Returns:
            pd.DataFrame: The API GET as a pandas DataFrames from Genesys.
        """
        genesys = Genesys(
            report_name=report_name,
            credentials=credentials_genesys,
            environment=environment,
            schedule_id=schedule_id,
            report_url=report_url,
            report_columns=report_columns,
        )

        df = genesys.to_df()

        logger.info(f"Downloaded the data from the Genesys into the Data Frame.")
        return df



import datetime
from typing import Any, Dict, List, Literal

import pandas as pd
from O365 import Account

from ..config import local_config
from ..exceptions import CredentialError
from .base import Source
import prefect


class Outlook(Source):
    def __init__(
        self,
        mailbox_name: str,
        start_date: str = None,
        end_date: str = None,
        credentials: Dict[str, Any] = None,
        limit: int = 10000,
        mailbox_folders: Literal[
            "sent", "inbox", "junk", "deleted", "drafts", "outbox", "archive"
        ] = ["sent", "inbox", "junk", "deleted", "drafts", "outbox", "archive"],
        request_retries: int = 10,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """
        Outlook connector build for fetching Outlook API source.
        Data are fetched from start to end date range. If start or end date are not provided
        then flow fetched data from yestarday by default.

        Args:
            mailbox_name (str): Mailbox name.
            start_date (str, optional): A filtering start date parameter e.g. "2022-01-01". Defaults to None.
            end_date (str, optional): A filtering end date parameter e.g. "2022-01-02". Defaults to None.
            credentials (Dict[str, Any], optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Application.
            Defaults to None.
            limit (int, optional): Number of fetched top messages. Defaults to 10000.
            mailbox_folders (Literal["sent", "inbox", "junk", "deleted", "drafts", "outbox", "archive"]):
                List of folders to select from the mailbox.  Defaults to ["sent", "inbox", "junk", "deleted", "drafts", "outbox", "archive"]
            request_retries (int, optional): How many times retries to authorizate. Defaults to 10.
        """

        self.logger = prefect.context.get("logger")

        self.credentials = credentials

        if self.credentials is None:
            raise CredentialError("You do not provide credentials!")

        self.request_retries = request_retries
        self.mailbox_name = mailbox_name
        self.start_date = start_date
        self.end_date = end_date
        if self.start_date is not None and self.end_date is not None:
            self.date_range_end_time = datetime.datetime.strptime(
                self.end_date, "%Y-%m-%d"
            )
            self.date_range_start_time = datetime.datetime.strptime(
                self.start_date, "%Y-%m-%d"
            )
        else:
            self.date_range_end_time = datetime.date.today() - datetime.timedelta(
                days=1
            )
            self.date_range_start_time = datetime.date.today() - datetime.timedelta(
                days=2
            )
            min_time = datetime.datetime.min.time()
            self.date_range_end_time = datetime.datetime.combine(
                self.date_range_end_time, min_time
            )
            self.date_range_start_time = datetime.datetime.combine(
                self.date_range_start_time, min_time
            )

        self.account = Account(
            (self.credentials["client_id"], self.credentials["client_secret"]),
            auth_flow_type="credentials",
            tenant_id=self.credentials["tenant_id"],
            main_resource=self.mailbox_name,
            request_retries=self.request_retries,
        )
        if self.account.authenticate():
            self.logger.info(f"{self.mailbox_name} Authenticated!")
        else:
            self.logger.info(f"{self.mailbox_name} NOT Authenticated!")

        self.mailbox_obj = self.account.mailbox()

        allowed_mailbox_folders = [
            "sent",
            "inbox",
            "junk",
            "deleted",
            "drafts",
            "outbox",
            "archive",
        ]
        if all(x in allowed_mailbox_folders for x in mailbox_folders):
            self.mailbox_folders = mailbox_folders
        else:
            raise Exception("Provided mailbox folders are not correct.")
        self.limit = limit

        super().__init__(*args, credentials=self.credentials, **kwargs)

    def to_df(self) -> pd.DataFrame:
        """Download Outlook data into a pandas DataFrame.

        Returns:
            pd.DataFrame: the DataFrame with time range.
        """
        data = []
        mailbox_generators_list = []
        for m in self.mailbox_folders:
            base_str = f"self.mailbox_obj.{m}_folder().get_messages({self.limit})"
            mailbox_generators_list.append(eval(base_str))

        for mailbox_generator in mailbox_generators_list:
            while True:
                try:
                    message = next(mailbox_generator)
                    received_time = message.received
                    date_time_str = str(received_time)
                    date_time_stripped = date_time_str[0:19]
                    date_obj = datetime.datetime.strptime(
                        date_time_stripped, "%Y-%m-%d %H:%M:%S"
                    )
                    if (
                        date_obj < self.date_range_start_time
                        or date_obj > self.date_range_end_time
                    ):
                        continue
                    else:
                        fetched = message.to_api_data()
                        try:
                            sender_mail = fetched["from"]["emailAddress"]["address"]
                            recivers_list = fetched.get("toRecipients")
                            recivers = " "
                            if recivers_list is not None:
                                recivers = ", ".join(
                                    reciver["emailAddress"]["address"]
                                    for reciver in recivers_list
                                )

                            categories = " "
                            if message.categories is not None:
                                categories = ", ".join(
                                    categories for categories in message.categories
                                )

                            conversation_index = " "
                            if message.conversation_index is not None:
                                conversation_index = message.conversation_index
                            row = {
                                "conversation ID": fetched.get("conversationId"),
                                "conversation index": conversation_index,
                                "categories": categories,
                                "sender": sender_mail,
                                "recivers": recivers,
                                "received_time": fetched.get("receivedDateTime"),
                            }

                            row["mail_adress"] = (
                                self.mailbox_name.split("@")[0]
                                .replace(".", "_")
                                .replace("-", "_")
                            )
                            if sender_mail == self.mailbox_name:
                                row["Inbox"] = False
                            else:
                                row["Inbox"] = True

                            data.append(row)
                        except KeyError as e:
                            self.logger.info("KeyError : " + str(e))
                except StopIteration:
                    break
        df = pd.DataFrame(data=data)

        return df

    def to_csv(self):
        df = self.to_df()
        file_name = self.mailbox_name.split("@")[0].replace(".", "_").replace("-", "_")
        df.to_csv(f"{file_name}.csv", index=False)



import paramiko
import prefect
from typing import Any, Dict, List
from io import StringIO, BytesIO
import prefect
import pandas as pd
from viadot.sources.base import Source
from pathlib import Path
import time
import os
from stat import S_ISDIR
from collections import defaultdict
from ..exceptions import CredentialError
import itertools


class SftpConnector(Source):
    def __init__(
        self,
        file_name: str = None,
        credentials_sftp: Dict[str, Any] = None,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """
        STFP connector which allows for download data files, listing and downloading into Data Frame.

        Args:
            file_name (str, optional): File name to download. Defaults to None.
            credentials_sftp (Dict[str, Any], optional): SFTP server credentials. Defaults to None.

        Notes:
            self.conn is paramiko.SFTPClient.from_transport method that contains additional methods like
            get, put, open etc. Some of them were not implemented in that class.
            For more check documentation (https://docs.paramiko.org/en/stable/api/sftp.html)

            sftp = SftpConnector()
            sftp.conn.open(filename='folder_a/my_file.zip', mode='r')


        Raises:
            CredentialError: If credentials are not provided in local_config or directly as a parameter.
        """

        self.logger = prefect.context.get("logger")

        self.credentials_sftp = credentials_sftp

        if self.credentials_sftp is None:
            raise CredentialError("Credentials not found.")

        self.file_name = file_name
        self.conn = None
        self.hostname = self.credentials_sftp.get("HOSTNAME")
        self.username = self.credentials_sftp.get("USERNAME")
        self.password = self.credentials_sftp.get("PASSWORD")
        self.port = self.credentials_sftp.get("PORT")
        self.rsa_key = self.credentials_sftp.get("RSA_KEY")

        self.recursive_files_list = {}
        self.file_name_list = []
        self.recursive_files = []

    def get_conn(self) -> paramiko.SFTPClient:
        """Returns a SFTP connection object.

        Returns: paramiko.SFTPClient.
        """

        ssh = paramiko.SSHClient()

        if self.conn is None:
            if self.rsa_key is None:
                transport = paramiko.Transport((self.hostname, self.port))
                transport.connect(None, self.username, self.password)

                self.conn = paramiko.SFTPClient.from_transport(transport)

            else:
                keyfile = StringIO(self.rsa_key)
                mykey = paramiko.RSAKey.from_private_key(keyfile)
                ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                ssh.connect(self.hostname, username=self.username, pkey=mykey)
                time.sleep(1)
                self.conn = ssh.open_sftp()

            return self.conn

    def get_cwd(self) -> str:
        """Return the current working directory for SFTP session.

        Returns:
            str: current working directory.
        """
        return self.conn.getcwd()

    def getfo_file(self, file_name: str) -> BytesIO:
        """Copy a remote file from the SFTP server and write to a file-like object.

        Args:
            file_name (str, optional): File name to copy.

        Returns:
            BytesIO: file-like object.
        """
        flo = BytesIO()
        try:
            self.conn.getfo(file_name, flo)
            return flo
        except Exception as e:
            self.logger.info(e)

    def to_df(
        self, file_name: str, sep: str = "\t", columns: List[str] = None, **kwargs
    ) -> pd.DataFrame:
        """Copy a remote file from the SFTP server and write it to Pandas dataframe.

        Args:
            file_name (str, optional): File name to download.
            sep (str, optional): The delimiter for the source file. Defaults to "\t".
            columns (List[str], optional): List of columns to select from file. Defaults to None.

        Returns:
            pd.DataFrame: Pandas dataframe.
        """
        byte_file = self.getfo_file(file_name=file_name)
        byte_file.seek(0)

        if Path(file_name).suffix == ".csv":
            df = pd.read_csv(byte_file, sep=sep, usecols=columns, **kwargs)

        elif Path(file_name).suffix == ".parquet":
            df = pd.read_parquet(byte_file, usecols=columns, **kwargs)

        elif Path(file_name).suffix == ".tsv":
            df = pd.read_csv(byte_file, sep="\t", usecols=columns, **kwargs)

        elif Path(file_name).suffix in [".xls", ".xlsx", ".xlsm"]:
            df = pd.read_excel(byte_file, usecols=columns, **kwargs)

        elif Path(file_name).suffix == ".json":
            df = pd.read_json(byte_file, **kwargs)

        elif Path(file_name).suffix == ".pkl":
            df = pd.read_pickle(byte_file, **kwargs)

        elif Path(file_name).suffix == ".sql":
            df = pd.read_sql(byte_file, **kwargs)

        elif Path(file_name).suffix == ".hdf":
            df = pd.read_hdf(byte_file, **kwargs)

        else:
            raise ValueError(
                f"Not able to read the file {Path(file_name).name}, unsupported filetype: {Path(file_name).suffix}"
            )

        return df

    def get_exported_files(self) -> List[str]:
        """List only exported files in current working directory.

        Returns:
            List: List of exported files.
        """
        self.file_name_list.clear()

        directory_structure = self.conn.listdir()
        for file in directory_structure:
            if "##exported.tsv" in file:
                self.file_name_list.append(file)

        return self.file_name_list

    def list_directory(self, path: str = None) -> List[str]:
        """Returns a list of files on the remote system.

        Args:
            path (str, optional): full path to the remote directory to list. Defaults to None.

        Returns:
            List: List of files.

        """

        if path is None:
            files = self.conn.listdir()
        else:
            files = self.conn.listdir(path)
        return files

    def recursive_listdir(self, path=".", files=None) -> defaultdict(list):
        """Recursively returns a defaultdict of files on the remote system.

        Args:
            path (str, optional): full path to the remote directory to list. Defaults to None.
            files (any, optional): parameter for calling function recursively.
        Returns:
            defaultdict(list): List of files.

        """

        if files is None:
            files = defaultdict(list)

        # loop over list of SFTPAttributes (files with modes)
        for attr in self.conn.listdir_attr(path):

            if S_ISDIR(attr.st_mode):
                # If the file is a directory, recurse it
                self.recursive_listdir(os.path.join(path, attr.filename), files)

            else:
                #  if the file is a file, add it to our dict
                files[path].append(attr.filename)

        self.recursive_files = files
        return files

    def process_defaultdict(self, defaultdict: any = None) -> list:
        """Process defaultdict to list of files.

        Args:
            defaultdict (any, optional): defaultdict of recursive files. Defaults to None.

        Returns:
            List: list of files.
        """
        path_list = []
        if defaultdict is None:
            defaultdict = self.recursive_files

        for item in list(defaultdict.items()):
            tuple_list_path = list(itertools.product([item[0]], item[1]))
            path_list.extend(
                [os.path.join(*tuple_path) for tuple_path in tuple_list_path]
            )
        return path_list

    def close_conn(self) -> None:
        """Closes the connection"""
        if self.conn is not None:
            self.conn.close()
            self.conn = None



from viadot.config import local_config
from viadot.sources import Outlook
import pandas as pd


def test_outlook_to_df():
    outlook_env_vars = local_config.get("OUTLOOK")
    outlook = Outlook(
        mailbox_name=outlook_env_vars["mail_example"],
        start_date="2022-04-28",
        end_date="2022-04-29",
    )
    df = outlook.to_df()
    assert isinstance(df, pd.DataFrame)



from .aselite import ASELiteToDF
from .azure_blob_storage import BlobFromCSV
from .azure_data_lake import (
    AzureDataLakeCopy,
    AzureDataLakeDownload,
    AzureDataLakeList,
    AzureDataLakeRemove,
    AzureDataLakeToDF,
    AzureDataLakeUpload,
)
from .azure_key_vault import (
    AzureKeyVaultSecret,
    CreateAzureKeyVaultSecret,
    DeleteAzureKeyVaultSecret,
)
from .azure_sql import (
    AzureSQLBulkInsert,
    AzureSQLCreateTable,
    AzureSQLDBQuery,
    AzureSQLToDF,
    AzureSQLUpsert,
    CheckColumnOrder,
    CreateTableFromBlob,
)
from .bcp import BCPTask
from .bigquery import BigQueryToDF
from .cloud_for_customers import C4CReportToDF, C4CToDF
from .github import DownloadGitHubFile
from .great_expectations import RunGreatExpectationsValidation
from .outlook import OutlookToDF
from .prefect_date_range import GetFlowNewDateRange
from .salesforce import SalesforceBulkUpsert, SalesforceToDF, SalesforceUpsert
from .sharepoint import SharepointToDF
from .sqlite import SQLiteInsert, SQLiteQuery, SQLiteSQLtoDF
from .supermetrics import SupermetricsToCSV, SupermetricsToDF
from .genesys import GenesysToCSV, GenesysToDF

try:
    from .sap_rfc import SAPRFCToDF
except ImportError:
    pass

from .duckdb import DuckDBCreateTableFromParquet, DuckDBQuery, DuckDBToDF
from .sql_server import SQLServerCreateTable, SQLServerToDF, SQLServerQuery

from .epicor import EpicorOrdersToDF

from .sftp import SftpToDF, SftpList
from .mindful import MindfulToCSV




from datetime import timedelta

import pandas as pd
from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs

try:
    from ..sources import SAPRFC
except ImportError:
    raise


class SAPRFCToDF(Task):
    def __init__(
        self,
        query: str = None,
        sep: str = None,
        func: str = None,
        rfc_total_col_width_character_limit: int = 400,
        credentials: dict = None,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        """
        A task for querying SAP with SQL using the RFC protocol.

        Note that only a very limited subset of SQL is supported:
        - aliases
        - where clauses combined using the AND operator
        - limit & offset

        Unsupported:
        - aggregations
        - joins
        - subqueries
        - etc.

        Args:
            query (str, optional): The query to be executed with pyRFC.
            sep (str, optional): The separator to use when reading query results. If not provided,
            multiple options are automatically tried. Defaults to None.
            func (str, optional): SAP RFC function to use. Defaults to None.
            rfc_total_col_width_character_limit (int, optional): Number of characters by which query will be split in chunks
            in case of too many columns for RFC function. According to SAP documentation, the limit is
            512 characters. However, we observed SAP raising an exception even on a slightly lower number
            of characters, so we add a safety margin. Defaults to 400.
            credentials (dict, optional): The credentials to use to authenticate with SAP.
            By default, they're taken from the local viadot config.
        """
        self.query = query
        self.sep = sep
        self.credentials = credentials
        self.func = func
        self.rfc_total_col_width_character_limit = rfc_total_col_width_character_limit

        super().__init__(
            name="sap_rfc_to_df",
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    @defaults_from_attrs(
        "query",
        "sep",
        "func",
        "rfc_total_col_width_character_limit",
        "credentials",
        "max_retries",
        "retry_delay",
    )
    def run(
        self,
        query: str = None,
        sep: str = None,
        credentials: dict = None,
        func: str = None,
        rfc_total_col_width_character_limit: int = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
    ) -> pd.DataFrame:
        """Task run method.

        Args:
            query (str, optional): The query to be executed with pyRFC.
            sep (str, optional): The separator to use when reading query results. If not provided,
            multiple options are automatically tried. Defaults to None.
            func (str, optional): SAP RFC function to use. Defaults to None.
            rfc_total_col_width_character_limit (int, optional): Number of characters by which query will be split in chunks
                in case of too many columns for RFC function. According to SAP documentation, the limit is
                512 characters. However, we observed SAP raising an exception even on a slightly lower number
                of characters, so we add a safety margin. Defaults to None.
        """
        if query is None:
            raise ValueError("Please provide the query.")
        sap = SAPRFC(
            sep=sep,
            credentials=credentials,
            func=func,
            rfc_total_col_width_character_limit=rfc_total_col_width_character_limit,
        )
        sap.query(query)
        self.logger.info(f"Downloading data from SAP to a DataFrame...")
        self.logger.debug(f"Running query: \n{query}.")

        df = sap.to_df()

        self.logger.info(f"Data has been downloaded successfully.")
        return df



import os
from typing import Any, Dict, List, Literal
import pandas as pd

from datetime import datetime
from prefect import Flow, task
from prefect.utilities import logging
from prefect.engine.signals import FAIL
from prefect.triggers import all_successful
from viadot.tasks import MindfulToCSV
from viadot.tasks import AzureDataLakeUpload
from viadot.task_utils import add_ingestion_metadata_task, adls_bulk_upload

logger = logging.get_logger()
file_to_adls_task = AzureDataLakeUpload()


@task
def add_timestamp(files_names: List = None, sep: str = "\t") -> None:
    """Add new column _viadot_downloaded_at_utc into each file given in the function.

    Args:
        files_names (List, optional): File names where to add the new column. Defaults to None.
        sep (str, optional): Separator type to load and to save data. Defaults to "\t".
    """
    if not files_names:
        logger.warning("Avoided adding a timestamp. No files were reported.")
    else:
        for file in files_names:
            df = pd.read_csv(file, sep=sep)
            df_updated = add_ingestion_metadata_task.run(df)
            df_updated.to_csv(file, index=False, sep=sep)


class MindfulToADLS(Flow):
    def __init__(
        self,
        name: str,
        credentials_mindful: Dict[str, Any] = None,
        credentials_secret: str = None,
        start_date: datetime = None,
        end_date: datetime = None,
        date_interval: int = 1,
        region: Literal["us1", "us2", "us3", "ca1", "eu1", "au1"] = "eu1",
        file_extension: Literal["parquet", "csv"] = "csv",
        sep: str = "\t",
        timeout: int = 3600,
        file_path: str = "",
        adls_file_path: str = None,
        adls_overwrite: bool = True,
        adls_sp_credentials_secret: str = None,
        *args: List[any],
        **kwargs: Dict[str, Any]
    ):
        """Mindful flow to download the CSV files and upload them to ADLS.

        Args:
            name (str): The name of the Flow.
            credentials_mindful (Dict[str, Any], optional): Credentials to connect with Mindful API. Defaults to None.
            credentials_secret (str, optional): Name of the credential secret to retreave the credentials. Defaults to None.
            start_date (datetime, optional): Start date of the request. Defaults to None.
            end_date (datetime, optional): End date of the resquest. Defaults to None.
            date_interval (int, optional): How many days are included in the request.
                If end_date is passed as an argument, date_interval will be invalidated. Defaults to 1.
            region (Literal[us1, us2, us3, ca1, eu1, au1], optional): SD region from where to interact with the mindful API. Defaults to "eu1".
            file_extension (Literal[parquet, csv], optional): File extensions for storing responses. Defaults to "csv".
            sep (str, optional): Separator in csv file. Defaults to "\t".
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
            file_path (str, optional): Path where to save the file locally. Defaults to ''.
            adls_file_path (str, optional): The destination path at ADLS. Defaults to None.
            adls_overwrite (bool, optional): Whether to overwrite files in the data lake. Defaults to True.
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
                ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET). Defaults to None.
        """

        self.credentials_mindful = credentials_mindful
        self.credentials_secret = credentials_secret
        self.start_date = start_date
        self.end_date = end_date
        self.date_interval = date_interval
        self.region = region
        self.file_extension = file_extension
        self.sep = sep
        self.file_path = file_path
        self.timeout = timeout

        self.adls_file_path = adls_file_path
        self.adls_overwrite = adls_overwrite
        self.adls_sp_credentials_secret = adls_sp_credentials_secret

        super().__init__(*args, name=name, **kwargs)

        self.mind_flow()

    def mind_flow(self) -> Flow:
        to_csv = MindfulToCSV(timeout=self.timeout)

        file_names = to_csv.bind(
            credentials_mindful=self.credentials_mindful,
            credentials_secret=self.credentials_secret,
            start_date=self.start_date,
            end_date=self.end_date,
            date_interval=self.date_interval,
            region=self.region,
            file_extension=self.file_extension,
            file_path=self.file_path,
            flow=self,
        )

        add_timestamp.bind(file_names, sep=self.sep, flow=self)

        adls_bulk_upload.bind(
            file_names=file_names,
            file_name_relative_path=self.file_path,
            adls_file_path=self.adls_file_path,
            adls_sp_credentials_secret=self.adls_sp_credentials_secret,
            adls_overwrite=self.adls_overwrite,
            timeout=self.timeout,
            flow=self,
        )

        add_timestamp.set_upstream(file_names, flow=self)
        adls_bulk_upload.set_upstream(add_timestamp, flow=self)
